#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file wifi_neighbor_node_list.py
 @brief This module gets IP routing table information.
 @date $Date$

 @author 株式会社 東日本計算センター

"""
import sys
import subprocess
import os
import errno
import re
from logging import getLogger, StreamHandler
from logging import INFO as LOG_LV
from abc import ABCMeta, abstractproperty, abstractmethod

import WiFiNodeIF

sys.path.append(".")

# Setup logger
logger = getLogger(__name__)
handler = StreamHandler()
handler.setLevel(LOG_LV)
logger.setLevel(LOG_LV)
logger.addHandler(handler)


##
# @class NotSupportedPlatform
# @brief
#
class NotSupportedPlatform(Exception):
    pass

##
# @class ParseFailure
# @brief


class ParseFailure(Exception):
    pass

##
# @class OutOfRange
# @brief


class OutOfRange(Exception):
    pass

##
# @class WiFiNeighborNodeList
# @brief 近隣ノード情報を保持するコンテナクラスを作成するための抽象基底クラス
#
# 近隣ノード情報を保持するコンテナクラスを作成するための抽象基底クラス。
# create_object()メソッドで、OSを判別して適切なオブジェクトを返す。
#


class WiFiNeighborNodeList(object):

    __metaclass__ = ABCMeta

    _node_list = []

    ##
    # @brief スキャン結果。Wi-Fi近隣ノード情報プロパティ。
    # @return 近隣ノード情報
    # WiFiNodeIF.NeighborNode(mac_address, channel, signal_level, essid)
    @property
    def node_list(self):
        return self._node_list

    ##
    # @brief 近隣ノードをスキャンする抽象メソッド。
    # @exception NotImplementedError 継承したクラスでメソッド未実装の場合に発生する。
    @abstractmethod
    def scan(self):
        raise NotImplementedError()

    ##
    # @brief OSを判別して適切なコンテナオブジェクトを返すクラスメソッド。
    # @param self
    # @exception NotSupportedPlatform
    @staticmethod
    def create_object():
        logger.info("platform: " + sys.platform)

        # 対応プラットフォームを増やす場合はここを修正する
        if sys.platform.startswith('linux'):
            return WiFiNeighborNodeListForLinux()
        else:
            logger.critical("ERROR: NOT supported platform")
            raise NotSupportedPlatform()


##
# @class WiFiNeighborNodeListForLinux
# @brief WiFiNeighborNodeList抽象基底クラスを継承。Linuxプラットフォーム Wireless-Tools用。
#
# iwlist scan コマンドを発行して、最終的に必要な情報に加工して返す。
# 標準エラー出力は/dev/nullに捨てる。
#
class WiFiNeighborNodeListForLinux(WiFiNeighborNodeList):

    COMMAND = "sudo iwlist scan 2> /dev/null"
    LINEFEED = "\n"
    os.environ["LANG"] = "C"

    ##
    # @brief コンストラクタ。
    #
    # コンストラクタ。プラットフォームを判別して、LinuxでなければNoneを返して終了。
    #
    # @param self
    # @exception NotSupportedPlatform プラットフォームがLinuxではなかった場合に発生する。
    def __init__(self):
        if not sys.platform.startswith('linux'):
            logger.critical("ERROR: NOT supported platform")
            raise NotSupportedPlatform(UserWarning, "NOT supported platform")

    ##
    # @brief iwlist scan コマンドを発行して、最終的に必要な情報に加工して返す。
    #
    # iwlist scan コマンドを発行して、最終的に必要な情報に加工する。
    # MAC Address, Channel, Signal level, ESSID を返す。
    # WiFiNodeIF.idl 参照。
    #
    # @param self
    # @exceptin OSError Wi-Fi デバイスが無いか、無効になっている
    # @exceptin ParseFailure "iwlist scan" によるスキャン結果が想定と異なっている
    def scan(self):
        # 結果格納用リスト初期化
        del self._node_list[:]

        mac_address = None
        channel = None
        signal_level = None
        essid = None

        try:
            stdout = subprocess.check_output(self.COMMAND, shell=True)
        except subprocess.CalledProcessError:
            logger.error("ERROR: command execution failure")
            raise

        rows = stdout.split(self.LINEFEED)
        rows = filter(lambda d: len(d) > 0, rows)

        # 長さが0ならWi-Fiデバイスが存在しない (filterの結果、長さが0でもNoneにはならない)
        if len(rows) == 0:
            logger.error("ERROR: Wi-Fi device is not exist")
            raise OSError(errno.ENXIO, "Wi-Fi device is not exist")

        # 1レコードずつ正規表現でマッチングして必要な行だけを抽出する
        # MAC, Channel, Lv, ESSID の順に並んでいるのは確定。ESSIDの行を処理したら1ノード処理終了。
        # 正規表現オブジェクトへのコンパイルは余力があれば実装する
        for data in rows:

            if re.match("^.+Address: ", data, 0):
                mac_address = data.replace(
                    re.match("^.+Address: ", data, 0).group(), "", 1)
                if not (re.match("^(([0-9]|[A-F]){2}:){5}([0-9]|[A-F]){2}$", mac_address, 0)):
                    logger.error(
                        "ERROR: parse failure (MAC Address): " + str(data))
                    raise ParseFailure("MAC Address")
                continue

            if re.match("^ +Channel:.*", data, 0):
                try:
                    channel = int(
                        data.replace(re.match("^ +Channel:", data, 0).group(), "", 1))
                except:
                    logger.error(
                        "ERROR: parse failure (Channel): " + str(data))
                    raise ParseFailure("Channel")

                if not ((0 <= channel) and (channel <= 99)):
                    logger.error(
                        "ERROR: out of range (Channel): " + str(channel))
                    raise OutOfRange("Channel")

                continue

            if re.match("^ +.+Signal level=.*", data, 0):
                try:
                    signal_level = data.replace(
                        re.match("^ +.+level=", data, 0).group(), "", 1)
                    signal_level = int(signal_level.replace(
                        re.search(" dBm", signal_level, 0).group(), "", 1))  # $付けるとミスマッチ
                except:
                    logger.error(
                        "ERROR: parse failure (Signal Level): " + str(data))
                    raise ParseFailure("Signal Level")

                if not ((-999 <= signal_level) and (signal_level <= 999)):
                    logger.error(
                        "ERROR: out of range (Signal Level): " + str(signal_level))
                    raise OutOfRange("Signal Level")

                continue

            # 取得データ確認 (iwlist scan コマンドでは Hidden ESSID のノードは取得できない)
            if re.match("^ +ESSID:.*", data, 0):
                essid = data.replace(
                    re.match('^ +ESSID:\"', data, 0).group(), "", 1)
                essid = essid.replace(re.search('\"', essid, 0).group(), "", 1)
                if not (re.match("^.+$", essid, 0)):
                    logger.error("ERROR: parse failure (ESSID): " + str(essid))
                    raise ParseFailure("ESSID")

                if (mac_address == None or channel == None or signal_level == None or essid == None):
                    logger.error("ERROR: parse failure (None exists)")
                    raise ParseFailure("None exists")

                self._node_list.append(
                    WiFiNodeIF.NeighborNode(mac_address, channel, signal_level, essid))

                mac_address = None
                channel = None
                signal_level = None
                essid = None

        if len(self._node_list) == 0:
            logger.error("ERROR: parse failure (Unexpected data)")
            raise ParseFailure("Unexpected data")
        

        return None
